/*****************************************************************************
 *
 * sxInit.c -- Initialise Starlight Xpress filter wheels
 *
 * Copyright 2014,2015 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <hidapi.h>

#include <errno.h>
#include <openastro/filterwheel.h>

#include "oafwprivate.h"
#include "unimplemented.h"
#include "sxfw.h"

#define	wheelState		wheel->_sx


static void _sxInitFunctionPointers ( oaFilterWheel* );
static int  _getNumSlots ( oaFilterWheel* );

/**
 * Initialise a given filter wheel device
 */

oaFilterWheel*
oaSXInitFilterWheel ( oaFilterWheelDevice* device )
{
  int					deviceAddr, deviceBus;
  hid_device*				hidHandle = 0;
  oaFilterWheel*			wheel;

  if (!( wheel = ( oaFilterWheel* ) malloc ( sizeof ( oaFilterWheel )))) {
    perror ( "malloc oaFilterWheel failed" );
    return 0;
  }

  OA_CLEAR ( wheelState );
  OA_CLEAR ( wheel->features );


  pthread_mutex_init ( &wheelState.hidMutex, 0 );

  ( void ) strcpy ( wheel->deviceName, device->deviceName );
  ( void ) strcpy ( wheelState.devicePath, device->sysPath );
  wheelState.initialised = 0;
  wheelState.index = -1;

  deviceAddr = device->_devIndex & 0xff;
  deviceBus = device->_devIndex >> 8;
  if (!( hidHandle = hid_open ( STARLIGHT_XPRESS_VID,
      STARLIGHT_XPRESS_FILTERWHEEL_PID, 0, deviceBus, deviceAddr ))) {
    fprintf ( stderr, "No matching USB device found!\n" );
    free ( wheel );
    return 0;
  }

  wheel->interface = device->interface;
  wheelState.hidHandle = hidHandle;
  wheelState.index = device->_devIndex;
  wheelState.wheelType = device->_devType;
  wheelState.currentPosition = 1;

  if (( wheelState.numSlots = _getNumSlots ( wheel )) < 1 ) {
    fprintf ( stderr, "%s: invalid number of slots in filter wheel\n",
        __FUNCTION__ );
    return 0;
  }

  _oaInitFilterWheelFunctionPointers ( wheel );
  _sxInitFunctionPointers ( wheel );

  wheel->features.coldReset = 0;
  wheel->features.warmReset = 0;
  wheel->features.speedControl = 0;

  return wheel;
}


static void
_sxInitFunctionPointers ( oaFilterWheel* wheel )
{
  wheel->funcs.initWheel = oaSXInitFilterWheel;
  wheel->funcs.getName = oaSXWheelGetName;
  wheel->funcs.closeWheel = oaSXWheelClose;
  // wheel->funcs.moveForward = _moveForward;
  // wheel->funcs.moveBackward = _moveBackward;
  wheel->funcs.moveTo = oaSXMoveTo;
  // wheel->funcs.warmReset = oaSXWheelWarmReset;
  // wheel->funcs.coldReset = oaSXWheelColdReset;
  wheel->funcs.numSlots = oaSXNumSlots;
}


int
_getNumSlots ( oaFilterWheel* wheel )
{
  unsigned char buffer[2];
  int numSlots;

  pthread_mutex_lock ( &wheelState.hidMutex );

  buffer[0] = 0;
  buffer[1] = 1;
  if ( _sxWheelWrite ( wheel, buffer )) {
    fprintf ( stderr, "%s: write error on get filters command\n",
      __FUNCTION__ );
    pthread_mutex_unlock ( &wheelState.hidMutex );
    return 0;
  }

  if ( _sxWheelRead ( wheel, buffer )) {
    pthread_mutex_unlock ( &wheelState.hidMutex );
    return 0;
  }

  do {
    sleep ( 1 );
    buffer[0] = 0;
    buffer[1] = 0;
    if ( _sxWheelWrite ( wheel, buffer )) {
      fprintf ( stderr, "%s: write error on get filters command\n",
        __FUNCTION__ );
      pthread_mutex_unlock ( &wheelState.hidMutex );
      return 0;
    }

    if ( _sxWheelRead ( wheel, buffer )) {
      pthread_mutex_unlock ( &wheelState.hidMutex );
      return 0;
    }
  } while ( !buffer[0] );

  pthread_mutex_unlock ( &wheelState.hidMutex );

  numSlots = buffer[1];
  return numSlots;
}


int
oaSXWheelClose ( oaFilterWheel* wheel )
{
  wheelState.initialised = 0;
  hid_close ( wheelState.hidHandle );
  return 0;
}
