/*****************************************************************************
 *
 * xagylControl-udev.c -- Xagyl filter wheel control functions (libudev)
 *
 * Copyright 2014,2015 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <errno.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <openastro/filterwheel.h>

#include "oafwprivate.h"
#include "unimplemented.h"
#include "xagylfw.h"

#define	wheelState		wheel->_xagyl


int
oaXagylGetWheelSpeed ( oaFilterWheel* wheel, unsigned int* speed )
{
  pthread_mutex_lock ( &wheelState.serialMutex );

  char buffer[50];
  int numRead;

  tcflush ( wheelState.fd, TCIFLUSH );

  if ( _xagylWheelWrite ( wheelState.fd, "i4", 2 )) {
    fprintf ( stderr, "%s: write error on command '%s'\n", __FUNCTION__,
        buffer );
    pthread_mutex_unlock ( &wheelState.serialMutex );
    return -1;
  }

  numRead = _xagylWheelRead ( wheelState.fd, buffer, 50 );
  if ( numRead > 0 ) {
    if ( strncmp ( buffer, "MaxSpeed ", 9 )) {
      fprintf ( stderr, "%s: '%s' failed to match expected string\n",
          __FUNCTION__, buffer );
      pthread_mutex_unlock ( &wheelState.serialMutex );
      return -1;
    }
    if ( sscanf ( buffer, "MaxSpeed %d%%\n", speed ) != 1 ) {
      buffer[ numRead ] = 0;
      fprintf ( stderr, "%s: wrong number of items(1) scanned from '%s'\n",
          __FUNCTION__, buffer );
      pthread_mutex_unlock ( &wheelState.serialMutex );
      return -1;
    }
  } else {
    fprintf ( stderr, "%s: no data read from wheel interface\n",
        __FUNCTION__ );
    pthread_mutex_unlock ( &wheelState.serialMutex );
    return -1;
  }

  pthread_mutex_unlock ( &wheelState.serialMutex );

  return 0;
}


int
oaXagylSetWheelSpeed ( oaFilterWheel* wheel, unsigned int speed, int nodelay )
{
  return 0;
}


int
oaXagylMoveTo ( oaFilterWheel* wheel, int slot, int nodelay )
{
  pthread_mutex_lock ( &wheelState.serialMutex );

  if ( nodelay ) {
    fprintf ( stderr, "%s: ignoring nodelay\n", __FUNCTION__ );
  }

  char buffer[50];
  char expected[50];
  int numRead;

  snprintf ( buffer, 3, "g%d", slot );
  snprintf ( expected, 3, "P%d", slot );

  tcflush ( wheelState.fd, TCIFLUSH );

  if ( _xagylWheelWrite ( wheelState.fd, buffer, 2 )) {
    fprintf ( stderr, "%s: write error on command '%s'\n", __FUNCTION__,
        buffer );
    pthread_mutex_unlock ( &wheelState.serialMutex );
    return -1;
  }

  numRead = _xagylWheelRead ( wheelState.fd, buffer, 50 );
  if ( numRead > 0 ) {
    if ( strncmp ( buffer, expected, 2 )) {
      fprintf ( stderr, "%s: '%s' failed to match expecting string '%s'\n",
          __FUNCTION__, buffer, expected );
      pthread_mutex_unlock ( &wheelState.serialMutex );
      return -1;
    }
  } else {
    fprintf ( stderr, "%s: no data read from wheel interface\n",
        __FUNCTION__ );
    pthread_mutex_unlock ( &wheelState.serialMutex );
    return -1;
  }

  pthread_mutex_unlock ( &wheelState.serialMutex );

  return 0;
}


int
oaXagylNumSlots ( oaFilterWheel* wheel )
{
  return wheelState.numSlots;
}
